@@ -2015,14 +2015,6 @@ static int fbcon_scroll(struct vc_data *vc, int t, int b, int dir,
 							(b - count)),
 				    vc->vc_video_erase_char,
 				    vc->vc_size_row * count);
-#ifdef CONFIG_UNICON
-			scr_memsetw((unsigned short *) (vc->vc_origin +
-							vc->vc_size_row * 
-							(b-count) +
-							vc->vc_screenbuf_size),
-				    vc->vc_video_erase_char & 0xff,
-				    vc->vc_size_row * count);
-#endif
 			return 1;
 		}
 		break;
@@ -2112,14 +2104,6 @@ static int fbcon_scroll(struct vc_data *vc, int t, int b, int dir,
 							t),
 				    vc->vc_video_erase_char,
 				    vc->vc_size_row * count);
-#ifdef CONFIG_UNICON
-			scr_memsetw((unsigned short *) (vc->vc_origin +
-							vc->vc_size_row *
-							t +
-							vc->vc_screenbuf_size),
-				   vc->vc_video_erase_char & 0xff,
-				   vc->vc_size_row * count);
-#endif
 			return 1;
 		}
 	}
